// (C) 2023-2024 by Folkert van Heusden <mail@vanheusden.com>
// CC0 license

#include <math.h>

#define INPUT_ANALOGUE_PB A0
// #define INPUT_ANALOGUE_BA A1  // balance

#define SAA_PIN_A0 12
#define SAA_PIN_NOT_WR 13
#define SAA_PIN_NOT_DTACK 10

#include <si5351.h>
#include <Wire.h>

Si5351 si5351;

uint8_t channels[6] { 255, 255, 255, 255, 255, 255 };

uint8_t rv[0x20] { };

void set_register(const uint8_t address, const uint8_t data) {
	rv[address] = data;

	digitalWrite(SAA_PIN_A0, HIGH);

	for(int i=0; i<8; i++)
		digitalWrite(2 + i, !!(address & (1 << i)));

	bool ack = digitalRead(SAA_PIN_NOT_DTACK);

	digitalWrite(SAA_PIN_NOT_WR, LOW);
	delayMicroseconds(5);
	digitalWrite(SAA_PIN_NOT_WR, HIGH);

	while(ack)
		ack = digitalRead(SAA_PIN_NOT_DTACK);

	delayMicroseconds(5);

	digitalWrite(SAA_PIN_A0, LOW);

	for(int i=0; i<8; i++)
		digitalWrite(2 + i, !!(data & (1 << i)));

	ack = digitalRead(SAA_PIN_NOT_DTACK);

	digitalWrite(SAA_PIN_NOT_WR, LOW);
	delayMicroseconds(5);
	digitalWrite(SAA_PIN_NOT_WR, HIGH);

	while(ack)
		ack = digitalRead(SAA_PIN_NOT_DTACK);
}

void reset_soundchip()
{
	memset(channels, 0xff, sizeof channels);

	memset(rv,       0x00, sizeof rv      );

	si5351.drive_strength(SI5351_CLK0, SI5351_DRIVE_6MA);
	si5351.set_freq(8000000ll * 100, SI5351_CLK0);

	set_register(0x1c, 2);  // reset

	for(uint8_t reg=0; reg<0x20; reg++)
		set_register(reg, 0);

	set_register(0x14, 63);  // channel frequency enable

	set_register(0x1c, 1);  // enable
}

void setup() {
	Serial.begin(115200);

	si5351.init(SI5351_CRYSTAL_LOAD_8PF, 0, 0);

	pinMode(SAA_PIN_NOT_WR, OUTPUT);  // /WR

	pinMode(SAA_PIN_A0, OUTPUT);  // A0 (control / address sel.)

	pinMode(SAA_PIN_NOT_DTACK, INPUT);  // /DTACK

	for(int i=2; i<10; i++)
		pinMode(i, OUTPUT);

	reset_soundchip();
}

static bool ungetc_valid = false;
static uint8_t ungetc_byte = 0x00;

void ungetc(const uint8_t c) {
	ungetc_byte = c;
	ungetc_valid = true;
}

constexpr const uint32_t base_freq = 2289;
constexpr const double   step      = (8000000ll - base_freq) / 1022;

uint8_t get_byte() {
	if (ungetc_valid) {
		ungetc_valid = false;

		return ungetc_byte;
	}

	while(Serial.available() == 0) {
		uint16_t v = analogRead(INPUT_ANALOGUE_PB);

		si5351.set_freq((v * step + base_freq) * SI5351_FREQ_MULT, SI5351_CLK0);
	}

	return Serial.read();
}

uint8_t allocate_channel(const uint8_t note)
{
	uint8_t empty = 255;

	for(uint8_t i=0; i<6; i++) {
		if (channels[i] == note)
			return i;

		if (channels[i] == 255)
			empty = i;
	}

	if (empty != 255)
		channels[empty] = note;

	return empty;
}

void free_channel(const uint8_t note)
{
	for(uint8_t i=0; i<6; i++) {
		if (channels[i] == note) {
			channels[i] = 255;
			break;
		}
	}
}

typedef struct {
	uint8_t midi_command, n_bytes : 2;
} mc_size_t;

const mc_size_t cmds[] = {
	{ 0x80, 2 },
	{ 0x90, 2 },
	{ 0xa0, 2 },
	{ 0xb0, 2 },
	{ 0xc0, 1 },
	{ 0xd0, 1 },
	{ 0xe0, 2 },
	{ 0xf0, 0 },
	{ 0, 0 }
};

bool midi_get(uint8_t *const cmd, uint8_t *const d1, uint8_t *const d2) {
	static uint8_t prev_cmd = 0x00;

	*cmd = get_byte();

	if (*cmd < 0x80) {
		if (!prev_cmd)
			return false;

		ungetc(*cmd);
		*cmd = prev_cmd;
	}

	const uint8_t cmdMasked = (*cmd) & 0xf0;

	*d1 = *d2 = 0x00;

	uint8_t i = 0;
	while(cmds[i].midi_command != cmdMasked && cmds[i].midi_command != 0x00)
		i++;

	if (cmds[i].midi_command == 0x00)
		return false;

	if (cmds[i].n_bytes >= 1) {
		*d1 = get_byte();

		if (*d1 & 128) {
			ungetc(*d1);
			*d1 = 0x00;
			return false;
		}
	}

	if (cmds[i].n_bytes >= 2) {
		*d2 = get_byte();

		if (*d2 & 128) {
			ungetc(*d2);
			*d2 = 0x00;
			return false;
		}
	}

	if (*cmd < 0xf0)
		prev_cmd = *cmd;

	return true;
}

void loop() {
	uint8_t cmd = 0;
	uint8_t d1  = 0;
	uint8_t d2  = 0;

	if (!midi_get(&cmd, &d1, &d2))
		return;

	if ((cmd & 0xf0) == 0x90 || (cmd & 0xf0) == 0x80) {
		if ((cmd & 0x0f) == 0x09)  // ignore percussion FIXME
			return;

		uint8_t ch = allocate_channel(d1);

		if (ch == 255)
			return;

		// set volume
		uint8_t volume   = d2 >> 3;  // midi is 7 bit volume, SAA1066 is 4 bit volume
		uint8_t v_reg_nr = ch;
		set_register(v_reg_nr, volume | (volume << 4));  // TODO panning

		if (d2 == 0)
			free_channel(d1);
		else {
			uint8_t octave   = (d1 + 1) / 12 - 1;

			// set octave
			uint8_t o_reg_nr = 0x10 + ch / 2;
			uint8_t new_val  = rv[o_reg_nr];
			if (ch & 1)
				new_val &= 0x0f, new_val |= octave << 4;
			else
				new_val &= 0xf0, new_val |= octave;
			set_register(o_reg_nr, new_val);


			// set note-frequency
			double frequency = pow(2, (d1 - 69.) / 12) * 440;
			// frequency = (15625 << octave) / (511 - offset)
			double temp      = 15625. / 4 * pow(2., octave + 1) / frequency;
			uint8_t note     = 511 - temp;
			uint8_t n_reg_nr = 0x08 + ch;
			set_register(n_reg_nr, note);
		}
	}
	else if (cmd == 0xff) {
		reset_soundchip();
	}
}
